// CHASECAM.H
// Description: Chase Cam Object
// Function:	Creates a Chase Cam object.  This code is Inspired
//              by the LokiCam object, created by Jeremy Blackman
//              at Monolith.  Which was, in turn, inspired by one
//              my tutorials.

#include "de_client.h"
#include "chasecam.h"

#include "windows.h"
#include "winbase.h"

#define CAMERA_MOVEMENT_LEFT        0
#define CAMERA_MOVEMENT_RIGHT       1
#define CAMERA_MOVEMENT_UP          2
#define CAMERA_MOVEMENT_ZOOM        3

// Constructor
ChaseCam::ChaseCam()
{
    VEC_SET(m_CameraPos,0.0f,0.0f,0.0f);
    m_CameraZoom = 100.0f;
    m_hCamera = DNULL;  
	m_CameraMode = 0;
}

// Destructor
ChaseCam::~ChaseCam()
{
    
}

// ----------------
// Init
//      Initializes the ChaseCam with a pointer to the client.
// ----------------
void ChaseCam::Init(ClientDE *pClient)
{
    m_pClientDE = pClient;
    VEC_SET(m_CameraPos,0.0f,0.0f,0.0f);

    // Sanity check
    if (m_pClientDE)
        m_pClientDE->SetupEuler(&m_CameraRotation,0.0f,0.0f,0.0f);

    m_CameraZoom = 100.0f;
    m_hCamera = DNULL;
	m_Offset = 38;
    m_Pitch = m_Yaw = m_Roll = 0.0f;
	m_CameraMode = 0;

    m_hTarget = pClient->GetClientObject();
}

// ----------------
// Update
//      Primary update loop.
// ----------------
void ChaseCam::Update()
{
    // Sanity check that we have a camera.
    if (m_hCamera)
    {
        // Sanity check that we have a target.
        if (!m_hTarget)
        {
            m_pClientDE->CPrint("Camera: No target set, defaulting to player.");
            m_hTarget = m_pClientDE->GetClientObject();
        }
        // Find appropriate sub-update to call.
        switch(m_CameraMode)
        {
            case 0:
                UpdateRelative();
                break;
            case 1:
                UpdateFreeform();
                break;
            default:
                m_pClientDE->CPrint("Camera: unknown camera mode!  Defaulting to relative.");
                m_CameraMode = 0;
                m_CameraZoom = 100.0f;
                UpdateRelative();
        }

    } else 
    {
        m_pClientDE->CPrint("Camera: *bzzzt* sorry, no camera set, thank you for playing.");
    }
}

// ----------------
// UpdateRelative
//      Updates the camera in a position behind the player, locked relative to them.
//      This mode was inspired by Ashley Matheson's tutorial.
// ----------------
void ChaseCam::UpdateRelative()
{
    DVector targetPos, cameraPos;
    DRotation targetRotation;
    DVector vUp, vRight, vForward, vOriginalForward;

    // Get some basic information about our target (probably the player)
    m_pClientDE->GetObjectPos(m_hTarget,&targetPos);
    m_pClientDE->GetObjectRotation(m_hTarget,&targetRotation);

    // Get the target's vectors
    m_pClientDE->GetRotationVectors(&targetRotation, &vUp, &vRight, &vForward); 
	vOriginalForward = vForward;

    // Since vForward is a unit vector, you can multiply by distance 
    // you want to move back..
    VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
    VEC_SUB(cameraPos, targetPos, vForward);
    cameraPos.y += m_Offset;

    m_pClientDE->SetObjectRotation(m_hCamera,&targetRotation);
    m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

    m_CameraPos = cameraPos;
    m_CameraRotation = targetRotation;
}


void ChaseCam::UpdateFreeform()
{
    DVector targetPos, cameraPos;
    DRotation cameraRotation;
    DVector vUp, vRight, vForward, vOriginalForward;

    // Get some basic information about our target (probably the player)
    m_pClientDE->GetObjectPos(m_hTarget,&targetPos);

    // Set up our rotation
    m_pClientDE->SetupEuler(&cameraRotation,m_Pitch,m_Yaw,m_Roll);

    // Get the camera's vectors
    m_pClientDE->GetRotationVectors(&cameraRotation, &vUp, &vRight, &vForward);
	vOriginalForward = vForward;

    // Since vForward is a unit vector, you can multiply by distance 
    // you want to move back..
    VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
    VEC_SUB(cameraPos, targetPos, vForward);
    cameraPos.y += m_Offset;

    m_pClientDE->SetObjectRotation(m_hCamera,&cameraRotation);
    m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

    m_CameraPos = cameraPos;
    m_CameraRotation = cameraRotation;
}

void ChaseCam::NextCamera()
{
	m_CameraMode++;
	if (m_CameraMode > MAXCAMERAS)
		m_CameraMode = 0;
}

// ----------------
// Zoom
//      Changes the camera's zoom by the amount in delta.
// ----------------
void ChaseCam::Zoom(float delta)
{
    m_CameraZoom += delta;

    if (m_CameraZoom <= 60.f) m_CameraZoom = 60.0f;
    if (m_CameraZoom >= 400.0f) m_CameraZoom = 400.0f;
}

// ----------------
// UpdateOffset
//      Updates the Offset (for freeform) by the amount in delta.
// ----------------
void ChaseCam::UpdateOffset(float delta)
{
    m_Offset += delta;

    if (m_Offset <= 0.0f) m_Offset = 0.0f;
    if (m_Offset >= 200.0f) m_Offset = 200.0f;
}

// ----------------
// UpdatePitch
//      Updates the Pitch (for freeform) by the amount in delta.
// ----------------
void ChaseCam::UpdatePitch(float delta)
{
    m_Pitch += delta;
}

// ----------------
// UpdateYaw
//      Updates the Yaw (for freeform) by the amount in delta.
// ----------------
void ChaseCam::UpdateYaw(float delta)
{
    m_Yaw += delta;
}

// ----------------
// UpdateRoll
//      Updates the Roll (for freeform) by the amount in delta.
// ----------------
void ChaseCam::UpdateRoll(float delta)
{
    m_Roll += delta;
}